#define _BSD_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <modbus/modbus.h>

typedef struct {
	modbus_param_t	mb_param;
	int		slave_id;

	GtkWidget       *window;

	GtkWidget	*device_entry;
	GtkWidget	*address_entry;

	GtkWidget	*link_status_label;
	GtkWidget	*current_label;
	GtkWidget	*voltage_label;
	GtkWidget	*potential_label;
	GtkWidget	*protecttime_label;
	GtkWidget	*mainsv_label;
	GtkWidget	*temp_label;
	GtkWidget	*mains_label;
	GtkWidget	*corrate1_label;
	GtkWidget	*corrate2_label;
	GtkWidget	*corrate3_label;
	GtkWidget	*datetime_label;
	GtkWidget	*rects1_label;
	GtkWidget	*rects2_label;
	GtkWidget	*controlstate_label;
	GtkWidget	*mode_combo;

	GtkWidget	*potential_entry;
	GtkWidget	*current_entry;
	
	GPtrArray	*active_buttons;
	GtkWidget	*connect;
} CpsControlApp;

CpsControlApp cpscontrolapp;
CpsControlApp *cpscontrol = &cpscontrolapp;
guint request_info_worker_id;

GKeyFile *conf;

gboolean request_info_worker(gpointer data)
{
	uint16_t	*buf;
	char		str[40];
	const char	na[]="N/A";
	int	start, end, i;
	start = 2000;
	end = 2017;

	buf = malloc( sizeof(*buf) * (end-start+1) );

	if (read_holding_registers( &cpscontrol->mb_param,
                                    cpscontrol->slave_id, start, end-start+1, buf ) < 0) {
		free( buf );
		gtk_label_set_text(GTK_LABEL( cpscontrol->link_status_label ), "No link");
		return TRUE;
	}
	gtk_label_set_text(GTK_LABEL( cpscontrol->link_status_label ), "Link OK");
	for (i=0; i<end-start+1; i++) {
		printf("%-8d0x%04X\t%8d\n", i+start, buf[i], (int16_t)buf[i]);
		if (i+start == 2000) { // current
			snprintf(str, sizeof(str), "%.1f", buf[i]/100.);
			gtk_label_set_text(GTK_LABEL( cpscontrol->current_label ), str);
		} else if (i+start == 2001) { // voltage
			snprintf(str, sizeof(str), "%.1f", buf[i]/10.);
			gtk_label_set_text(GTK_LABEL( cpscontrol->voltage_label ), str);
		} else if (i+start == 2002) { // potential
			snprintf(str, sizeof(str), "%.1f", ((int16_t)buf[i])/100.);
			gtk_label_set_text(GTK_LABEL( cpscontrol->potential_label ), str);
		} else if (i+start == 2003) { // protect time
			int pt = buf[i] | (int)buf[i+1]<<16;
			int hours, min;
			hours = pt/3600;
			pt -= hours*3600;
			min = pt/60;
			pt -= min*60;
			snprintf(str, sizeof(str), "%d:%d:%d\n", hours, min, pt);
			printf("Protection time: %s", str);
			gtk_label_set_text(GTK_LABEL( cpscontrol->protecttime_label ), str);
		} else if (i+start == 2005) { // mains voltage
			if (buf[i])
				snprintf(str, sizeof(str), "%d", buf[i]);
			else
				strcpy(str, na);
			gtk_label_set_text(GTK_LABEL( cpscontrol->mainsv_label ), str);
		} else if (i+start == 2006) { // temperature
			if ((int16_t)buf[i] != -100) // sensor error
				snprintf(str, sizeof(str), "%d", (int16_t)buf[i]);
			else
				strcpy(str, na);
			gtk_label_set_text(GTK_LABEL( cpscontrol->temp_label ), str);
		} else if (i+start == 2007 || i+start == 2008) { // rectifiers state
			int j;
			str[0]='\0';
			for (j=0; j<8; j++) {
				switch ((buf[i] >> j*2) & 0x03) {
				case 0:
					strcat(str, "ON ");
					break;
				case 1:
					strcat(str, "OF ");
					break;
				case 2:
					strcat(str, "-- ");
					break;
				case 3:
					strcat(str, "AL ");
					break;
				}
			}
			gtk_label_set_text(GTK_LABEL(
						   i+start == 2007 ? cpscontrol->rects1_label
						   : cpscontrol->rects2_label), str);
		} else if (i+start == 2009) { // sensors state
			int j;
			for (j=0; j<4; j++)
				printf(((buf[i] >> j) & 1) ? "+":"-");
			if (buf[i] & 1)
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate1_label ), "1 fail");
			else
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate1_label ), "1 ok");
			if (buf[i] & 2)
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate2_label ), "2 fail");
			else
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate2_label ), "2 ok");
			if (buf[i] & 4)
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate3_label ), "3 fail");
			else
				gtk_label_set_text(GTK_LABEL( cpscontrol->corrate3_label ), "3 ok");
			
			printf(" Mains: ");
			switch (buf[i]>>4 & 0x03) {
			case 0:
				strcpy(str, "ok");
				break;
			case 1:
				strcpy(str, "none");
				break;
			case 2:
				strcpy(str, "lower");
				break;
			case 3:
				strcpy(str, "higher");
			}
			printf("%s\n", str);
			gtk_label_set_text(GTK_LABEL( cpscontrol->mains_label ), str);
		} else if (i+start == 2010) { // time
			long int t=buf[i] | (uint64_t)buf[i+1]<<16 |
				(uint64_t)buf[i+2]<<32 | (uint64_t)buf[i+3]<<48;
			printf("Local time: %s\n", ctime(&t));
			gtk_label_set_text(GTK_LABEL( cpscontrol->datetime_label ), ctime(&t));
		} else if (i+start == 2017) { // mode
			switch (buf[i]) {
			case 0x00: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "MC output current");
				break;
			case 0x01: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "MC output voltage");
				break;
			case 0x02: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "MC polarisation potential");
				break;
			case 0x03: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "RC output current");
				break;
			case 0x05: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "RC polarisation potential");
				break;
			case 0x06: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "MC full potential");
				break;
			case 0x07: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "RC full potential");
				break;
			case 0x08: gtk_label_set_text(GTK_LABEL( cpscontrol->controlstate_label ),
						      "Waiting");
				break;
			}
		}
	}
	free( buf );
	return TRUE;
}

void set_button_sensitivity(gpointer data, gpointer user_data)
{
	gtk_widget_set_sensitive(GTK_WIDGET( data ), (gboolean)user_data);
}

void 
on_window_destroy (GtkObject *object, gpointer user_data)
{
	g_key_file_set_string(conf, "General", "Device",
			      gtk_entry_get_text(GTK_ENTRY(cpscontrol->device_entry)));
	g_key_file_set_string(conf, "General", "Address",
			      gtk_entry_get_text(GTK_ENTRY(cpscontrol->address_entry)));
	gtk_main_quit ();
}

void 
on_connect_button_clicked (GtkObject *object, gpointer user_data)
{
    	modbus_init_rtu( &cpscontrol->mb_param,
			 gtk_entry_get_text(GTK_ENTRY( cpscontrol->device_entry )), 9600, "none", 8, 1 );
#if 1
	modbus_set_debug( &cpscontrol->mb_param, 1 );
#endif
	if (modbus_connect( &cpscontrol->mb_param ) != -1) {
		gtk_widget_set_sensitive(GTK_WIDGET( cpscontrol->connect ), FALSE);
		g_ptr_array_foreach( cpscontrol->active_buttons, set_button_sensitivity, (gpointer)TRUE );
		request_info_worker_id = g_idle_add(request_info_worker, NULL);
	}
}

void 
on_disconnect_button_clicked (GtkObject *object, gpointer user_data)
{
	g_source_remove( request_info_worker_id );
	modbus_close( &cpscontrol->mb_param );
	gtk_widget_set_sensitive(GTK_WIDGET( cpscontrol->connect ), TRUE);
	g_ptr_array_foreach( cpscontrol->active_buttons, set_button_sensitivity, FALSE );
}

void 
on_address_entry_changed (GtkObject *object, gpointer user_data)
{
	cpscontrol->slave_id = atoi( gtk_entry_get_text(GTK_ENTRY( object )) );
}

void
on_setpotential_button_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &cpscontrol->mb_param,
				cpscontrol->slave_id,
				2016,
				strtod(gtk_entry_get_text(
					       GTK_ENTRY( cpscontrol->potential_entry )),
				       NULL)*100 );
}

void
on_setcurrent_button_clicked (GtkObject *object, gpointer user_data)
{
	preset_single_register( &cpscontrol->mb_param,
				cpscontrol->slave_id,
				2014,
				strtod(gtk_entry_get_text(
					       GTK_ENTRY( cpscontrol->current_entry )),
				       NULL)*100 );
}

void
on_setmode_button_clicked (GtkObject *object, gpointer user_data)
{
	gint a;
	int r;
	a = gtk_combo_box_get_active(GTK_COMBO_BOX( cpscontrol->mode_combo ));
	switch (a) {
	case 0: r=0x03; break;
	case 1: r=0x05; break;
	case 2: r=0x06; break;
	case 3: r=0x08; break;
	}
	preset_single_register( &cpscontrol->mb_param,
				cpscontrol->slave_id,
				2017, r );
}

void
on_setdatetime_button_clicked (GtkObject *object, gpointer user_data)
{
	uint16_t buf[4];
	long long int t=time(NULL);
	buf[0]=t&0xffff;
	buf[1]=(t>>16)&0xffff;
	buf[2]=(t>>32)&0xffff;
	buf[3]=(t>>48)&0xffff;
	preset_multiple_registers( &cpscontrol->mb_param,
				   cpscontrol->slave_id,
				   2010,
				   4,
				   buf );
}

int
main (int argc, char *argv[])
{
	GtkBuilder      *builder;
	gchar *conf_path, *conf_path_new, *conf_data;
	const gchar *tmp;
	int newconffile;

	setlocale(LC_ALL, "");
	
	gtk_init (&argc, &argv);
 
	builder = gtk_builder_new ();
	gtk_builder_add_from_file (builder, "cps_control.xml", NULL);
	cpscontrol->window = GTK_WIDGET (gtk_builder_get_object (
						  builder, "mainwindow"));
	cpscontrol->device_entry = GTK_WIDGET (gtk_builder_get_object (
							builder, "device_entry"));
	cpscontrol->address_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "address_entry"));

	cpscontrol->current_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "current_label"));
	cpscontrol->link_status_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "link_status_label"));
	cpscontrol->voltage_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "voltage_label"));
	cpscontrol->potential_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "potential_label"));
	cpscontrol->protecttime_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "protecttime_label"));
	cpscontrol->mainsv_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "mainsv_label"));
	cpscontrol->temp_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "temp_label"));
	cpscontrol->mains_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "mains_label"));
	cpscontrol->corrate1_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "corrate1_label"));
	cpscontrol->corrate2_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "corrate2_label"));
	cpscontrol->corrate3_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "corrate3_label"));
	cpscontrol->datetime_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "datetime_label"));
	cpscontrol->controlstate_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "controlstate_label"));
	cpscontrol->rects1_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "rects1_label"));
	cpscontrol->rects2_label = GTK_WIDGET (gtk_builder_get_object (
							 builder, "rects2_label"));
	cpscontrol->potential_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "potential_entry"));
	cpscontrol->current_entry = GTK_WIDGET (gtk_builder_get_object (
							 builder, "current_entry"));
	cpscontrol->mode_combo = GTK_WIDGET (gtk_builder_get_object (
							 builder, "mode_combo"));

	cpscontrol->connect = GTK_WIDGET ( gtk_builder_get_object (
						    builder, "connect_button"));
	cpscontrol->active_buttons = g_ptr_array_new();
	g_ptr_array_add( cpscontrol->active_buttons, gtk_builder_get_object(
				   builder, "disconnect_button" ));
	g_ptr_array_add( cpscontrol->active_buttons, gtk_builder_get_object(
				 builder, "setdatetime_button" ));
	g_ptr_array_add( cpscontrol->active_buttons, gtk_builder_get_object(
				   builder, "setcurrent_button" ));
	g_ptr_array_add( cpscontrol->active_buttons, gtk_builder_get_object(
				 builder, "setmode_button" ));
	g_ptr_array_add( cpscontrol->active_buttons, gtk_builder_get_object(
				   builder, "setpotential_button" ));

	gtk_builder_connect_signals (builder, NULL);
 
	g_object_unref (G_OBJECT (builder));

	conf = g_key_file_new();
	tmp = g_get_home_dir();
	conf_path = g_build_filename(tmp, ".cps_control", NULL);
	conf_path_new = g_build_filename(tmp, ".cps_control.new", NULL);
	/* Why the hell manual doesn't explain when and how they get freed */
	/* g_free(tmp); */
	if (g_key_file_load_from_file(conf, conf_path,
				  G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS,
				      NULL)==TRUE) {
		tmp = g_key_file_get_string(conf, "General", "Device", NULL);
		gtk_entry_set_text(GTK_ENTRY( cpscontrol->device_entry ), tmp);
		/* g_free(tmp); */
		tmp = g_key_file_get_string(conf, "General", "Address", NULL);
		gtk_entry_set_text(GTK_ENTRY( cpscontrol->address_entry ), tmp);
		/* g_free(tmp); */
	}
	
	gtk_widget_show (cpscontrol->window);                
	gtk_main ();

	conf_data = g_key_file_to_data(conf, NULL, NULL);
	newconffile = g_open(conf_path_new, O_WRONLY|O_TRUNC|O_CREAT, 0600);
	if (newconffile==-1)
		perror("Warning: can't save configuration data");
	else {
		write(newconffile, conf_data, strlen(conf_data));
#ifndef NATIVE_WIN32
		if (fsync(newconffile)==-1) {
			close(newconffile);
			perror("Warning: can't sync configuration data");
		} else {
			close(newconffile);
			g_rename(conf_path_new, conf_path);
		}
#else
		close(newconffile);
		g_rename(conf_path_new, conf_path);
#endif
	}

	g_free(conf_data);
	g_free(conf_path_new);
	g_free(conf_path);
	g_key_file_free(conf);
		
	return 0;
}

